@@ -391,6 +391,27 @@ LXWriteMSR(unsigned long addr, unsigned long lo, unsigned long hi)
     LX_MSR_WRITE(addr, lo, hi);
 }
 
+#ifdef XSERVER_LIBPCIACCESS
+static inline void * map_pci_mem(ScrnInfoPtr pScrni, int vram,
+				 struct pci_device *dev,
+				 int bar, int size)
+{
+  void *ptr;
+  void** result = (void**)&ptr;
+  int map_size = size ? size : dev->regions[bar].size;
+    
+  int err = pci_device_map_range(dev,
+				 dev->regions[bar].base_addr,
+				 map_size,
+				 PCI_DEV_MAP_FLAG_WRITABLE |
+				 (vram ? PCI_DEV_MAP_FLAG_WRITE_COMBINE : 0),
+				 result);
+  
+  if (err) 
+    return NULL;
+  return ptr;
+}
+#endif
 
 static Bool
 LXMapMem(ScrnInfoPtr pScrni)
@@ -398,9 +419,12 @@ LXMapMem(ScrnInfoPtr pScrni)
     GeodeRec *pGeode = GEODEPTR(pScrni);
     int index = pScrni->scrnIndex;
     unsigned long cmd_bfr_phys;
+
+    pciVideoPtr pci = xf86GetPciInfoForEntity(pGeode->pEnt->index);
+
+#ifndef XSERVER_LIBPCIACCESS
     PCITAG tag;
 
-    pciVideoRec *pci = xf86GetPciInfoForEntity(pGeode->pEnt->index);
     tag = pciTag(pci->bus, pci->device, pci->func);
 
     cim_gp_ptr = (unsigned char *)xf86MapPciMem(index, VIDMEM_MMIO,
@@ -417,18 +441,26 @@ LXMapMem(ScrnInfoPtr pScrni)
 
     cim_fb_ptr = (unsigned char *)xf86MapPciMem(index, VIDMEM_FRAMEBUFFER,
 						tag, pci->memBase[0], pGeode->FBAvail + CIM_CMD_BFR_SZ);
+#else
+    cim_gp_ptr = map_pci_mem(pScrni, 0, pci, 1, LX_GP_REG_SIZE);
+    cim_vg_ptr = map_pci_mem(pScrni, 0, pci, 2, LX_VG_REG_SIZE);
+    cim_vid_ptr = map_pci_mem(pScrni, 0, pci, 3, LX_VID_REG_SIZE);
+    cim_vip_ptr = map_pci_mem(pScrni, 0, pci, 4, LX_VIP_REG_SIZE);
+    cim_fb_ptr = map_pci_mem(pScrni, 1, pci, 0, pGeode->FBAvail + CIM_CMD_BFR_SZ);
+#endif
 
     if (pScrni->memPhysBase == 0)
-      pScrni->memPhysBase = pci->memBase[0];
+      pScrni->memPhysBase = PCI_REGION_BASE(pci, 0, REGION_MEM);
+      
 
-    cmd_bfr_phys = pci->memBase[0] + pGeode->CmdBfrOffset;
+    cmd_bfr_phys = PCI_REGION_BASE(pci, 0, REGION_MEM) + pGeode->CmdBfrOffset;
     cim_cmd_base_ptr = cim_fb_ptr + pGeode->CmdBfrOffset;
 
     if (!cim_gp_ptr || !cim_vg_ptr || !cim_vid_ptr || !cim_fb_ptr ||
 	!cim_vip_ptr)
       return FALSE;
 
-    gp_set_frame_buffer_base(pci->memBase[0], pGeode->FBAvail);
+    gp_set_frame_buffer_base(PCI_REGION_BASE(pci, 0, REGION_MEM), pGeode->FBAvail);
     gp_set_command_buffer_base(cmd_bfr_phys, 0, pGeode->CmdBfrSize);
 
     XpressROMPtr = xf86MapVidMem(index, VIDMEM_FRAMEBUFFER, 0xF0000, 0x10000);
@@ -551,6 +583,18 @@ LXPreInit(ScrnInfoPtr pScrni, int flags)
     if (!xf86SetDefaultVisual(pScrni, -1))
 	return FALSE;
 
+    /*
+     * If the driver can do gamma correction, it should call xf86SetGamma()
+     * here.
+     */
+    {
+        Gamma zeros = { 0.0, 0.0, 0.0 };
+
+        if (!xf86SetGamma(pScrni, zeros)) {
+            return FALSE;
+        }
+    }
+
     pScrni->progClock = TRUE;
     xf86CollectOptions(pScrni, NULL);
     xf86ProcessOptions(pScrni->scrnIndex, pScrni->options, GeodeOptions);
@@ -845,9 +889,10 @@ LXSetVideoMode(ScrnInfoPtr pScrni, DisplayModePtr pMode)
     lx_disable_dac_power(pScrni, DF_CRT_DISABLE);
     vg_set_compression_enable(0);
 
-    if (!pMode->type || pMode->type == M_T_USERDEF) 
-      lx_set_custom_mode(pGeode, pMode, pScrni->bitsPerPixel);
-    else {
+    /* If the mode is a default one, then set the mode with the Cimarron
+     * tables */
+
+    if ((pMode->type & M_T_BUILTIN) || (pMode->type & M_T_DEFAULT)) {
       if (pMode->Flags & V_NHSYNC)
 	flags |= VG_MODEFLAG_NEG_HSYNC;
       if (pMode->Flags & V_NVSYNC)
@@ -878,8 +923,14 @@ LXSetVideoMode(ScrnInfoPtr pScrni, DisplayModePtr pMode)
 			    pScrni->bitsPerPixel, GeodeGetRefreshRate(pMode), 
 			    0);
       }
-    } 
-   
+    }
+    else {
+	/* For anything other then a default mode - use the passed in
+	 * timings */
+
+	lx_set_custom_mode(pGeode, pMode, pScrni->bitsPerPixel);
+    }
+
     if (pGeode->Output & OUTPUT_PANEL)
       df_set_output_path((pGeode->Output & OUTPUT_CRT) ? DF_DISPLAY_CRT_FP : DF_DISPLAY_FP);
     else
@@ -1386,7 +1437,9 @@ LXValidMode(int scrnIndex, DisplayModePtr pMode, Bool Verbose, int flags)
 
     memset(&vgQueryMode, 0, sizeof(vgQueryMode));
 
-    if (pMode->type && pMode->type != M_T_USERDEF) {
+    /* For builtin and default modes, try to look up the mode in Cimarron */
+
+    if ((pMode->type & M_T_BUILTIN) || (pMode->type & M_T_DEFAULT)) {
       
       if (pGeode->Output & OUTPUT_PANEL) {
 
